#!/usr/bin/python
'''
Provides IPython console widget.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{http://www.opensource.org/licenses/bsd-license.php}

http://ipython.scipy.org/moin/Cookbook/EmbeddingInGTK
'''

import gtk, gobject
import re
import sys
import os
import pango
from StringIO import StringIO
import thread

try:
    import IPython
except ImportError:
    IPython = None

class IterableIPShell:
    '''
    Create an IPython instance. Does not start a blocking event loop,
    instead allow single iterations. This allows embedding in GTK+
    without blockage.

    @ivar IP: IPython instance.
    @type IP: IPython.iplib.InteractiveShell
    @ivar iter_more: Indicates if the line executed was a complete command,
    or we should wait for more.
    @type iter_more: integer
    @ivar history_level: The place in history where we currently are
    when pressing up/down.
    @type history_level: integer
    @ivar complete_sep: Seperation delimeters for completion function.
    @type complete_sep: _sre.SRE_Pattern
    '''
    def __init__(self,argv=[],user_ns=None,user_global_ns=None,
                 cin=None, cout=None,cerr=None, input_func=None):
        '''


        @param argv: Command line options for IPython
        @type argv: list
        @param user_ns: User namespace.
        @type user_ns: dictionary
        @param user_global_ns: User global namespace.
        @type user_global_ns: dictionary.
        @param cin: Console standard input.
        @type cin: IO stream
        @param cout: Console standard output.
        @type cout: IO stream
        @param cerr: Console standard error.
        @type cerr: IO stream
        @param input_func: Replacement for builtin raw_input()
        @type input_func: function
        '''
        if input_func:
            IPython.iplib.raw_input_original = input_func
        if cin:
            IPython.Shell.Term.cin = cin
        if cout:
            IPython.Shell.Term.cout = cout
        if cerr:
            IPython.Shell.Term.cerr = cerr

        # This is to get rid of the blockage that accurs during
        # IPython.Shell.InteractiveShell.user_setup()
        IPython.iplib.raw_input = lambda x: None

        self.term = IPython.genutils.IOTerm(cin=cin, cout=cout, cerr=cerr)
        os.environ['TERM'] = 'dumb'
        excepthook = sys.excepthook
        self.IP = IPython.Shell.make_IPython(
          argv,user_ns=user_ns,
          user_global_ns=user_global_ns,
          embedded=True,
          shell_class=IPython.Shell.InteractiveShell)
        self.IP.system = lambda cmd: self.shell(self.IP.var_expand(cmd),
                                                header='IPython system call: ',
                                                verbose=self.IP.rc.system_verbose)
        sys.excepthook = excepthook
        self.iter_more = 0
        self.history_level = 0
        self.complete_sep =  re.compile('[\s\{\}\[\]\(\)]')

    def execute(self):
        '''
        Executes the current line provided by the shell object.
        '''
        self.history_level = 0
        orig_stdout = sys.stdout
        sys.stdout = IPython.Shell.Term.cout
        try:
            line = self.IP.raw_input(None, self.iter_more)
            if self.IP.autoindent:
                self.IP.readline_startup_hook(None)
        except KeyboardInterrupt:
            self.IP.write('\nKeyboardInterrupt\n')
            self.IP.resetbuffer()
            # keep cache in sync with the prompt counter:
            self.IP.outputcache.prompt_count -= 1

            if self.IP.autoindent:
                self.IP.indent_current_nsp = 0
            self.iter_more = 0
        except:
            self.IP.showtraceback()
        else:
            self.iter_more = self.IP.push(line)
            if (self.IP.SyntaxTB.last_syntax_error and
                self.IP.rc.autoedit_syntax):
                self.IP.edit_syntax_error()
        if self.iter_more:
            self.prompt = str(self.IP.outputcache.prompt2).strip()
            if self.IP.autoindent:
                self.IP.readline_startup_hook(self.IP.pre_readline)
        else:
            self.prompt = str(self.IP.outputcache.prompt1).strip()
        sys.stdout = orig_stdout

    def historyBack(self):
        '''
        Provides one history command back.

        @return: The command string.
        @rtype: string
        '''
        self.history_level -= 1
        return self._getHistory()

    def historyForward(self):
        '''
        Provides one history command forward.

        @return: The command string.
        @rtype: string
        '''
        self.history_level += 1
        return self._getHistory()

    def _getHistory(self):
        '''
        Get's the command string of the current history level.

        @return: Historic command string.
        @rtype: string
        '''
        try:
            rv = self.IP.user_ns['In'][self.history_level].strip('\n')
        except IndexError:
            self.history_level = 0
            rv = ''
        return rv

    def updateNamespace(self, ns_dict):
        '''
        Add the current dictionary to the shell namespace.

        @param ns_dict: A dictionary of symbol-values.
        @type ns_dict: dictionary
        '''
        self.IP.user_ns.update(ns_dict)

    def complete(self, line):
        '''
        Returns an auto completed line and/or posibilities for completion.

        @param line: Given line so far.
        @type line: string

        @return: Line completed as for as possible,
        and possible further completions.
        @rtype: tuple
        '''
        split_line = self.complete_sep.split(line)
        possibilities = self.IP.complete(split_line[-1])
        if possibilities:
            def _commonPrefix(str1, str2):
                '''
                Reduction function. returns common prefix of two given strings.

                @param str1: First string.
                @type str1: string
                @param str2: Second string
                @type str2: string

                @return: Common prefix to both strings.
                @rtype: string
                '''
                for i in range(len(str1)):
                    if not str2.startswith(str1[:i+1]):
                        return str1[:i]
                return str1
            common_prefix = reduce(_commonPrefix, possibilities)
            completed = line[:-len(split_line[-1])]+common_prefix
        else:
            completed = line
        return completed, possibilities


    def shell(self, cmd,verbose=0,debug=0,header=''):
        '''
        Replacement method to allow shell commands without them blocking.

        @param cmd: Shell command to execute.
        @type cmd: string
        @param verbose: Verbosity
        @type verbose: integer
        @param debug: Debug level
        @type debug: integer
        @param header: Header to be printed before output
        @type header: string
        '''
        stat = 0
        if verbose or debug: print header+cmd
        # flush stdout so we don't mangle python's buffering
        if not debug:
            input, output = os.popen4(cmd)
            print output.read()
            output.close()
            input.close()

class ConsoleView(gtk.TextView):
    '''
    Specialized text view for console-like workflow.

    @cvar ANSI_COLORS: Mapping of terminal colors to X11 names.
    @type ANSI_COLORS: dictionary

    @ivar text_buffer: Widget's text buffer.
    @type text_buffer: gtk.TextBuffer
    @ivar color_pat: Regex of terminal color pattern
    @type color_pat: _sre.SRE_Pattern
    @ivar mark: Scroll mark for automatic scrolling on input.
    @type mark: gtk.TextMark
    @ivar line_start: Start of command line mark.
    @type line_start: gtk.TextMark
    '''
    ANSI_COLORS =  {'0;30': 'Black',     '0;31': 'Red',
                    '0;32': 'Green',     '0;33': 'Brown',
                    '0;34': 'Blue',      '0;35': 'Purple',
                    '0;36': 'Cyan',      '0;37': 'LightGray',
                    '1;30': 'DarkGray',  '1;31': 'DarkRed',
                    '1;32': 'SeaGreen',  '1;33': 'Yellow',
                    '1;34': 'LightBlue', '1;35': 'MediumPurple',
                    '1;36': 'LightCyan', '1;37': 'White'}

    def __init__(self):
        '''
        Initialize console view.
        '''
        gtk.TextView.__init__(self)
        self.modify_font(pango.FontDescription('Mono'))
        self.set_cursor_visible(True)
        self.text_buffer = self.get_buffer()
        self.mark = self.text_buffer.create_mark('scroll_mark',
                                                 self.text_buffer.get_end_iter(),
                                                 False)
        for code in self.ANSI_COLORS:
            self.text_buffer.create_tag(code,
                                        foreground=self.ANSI_COLORS[code],
                                        weight=700)
        self.text_buffer.create_tag('0')
        self.text_buffer.create_tag('notouch', editable=False)
        self.color_pat = re.compile('\x01?\x1b\[(.*?)m\x02?')
        self.line_start = \
            self.text_buffer.create_mark('line_start',
                                         self.text_buffer.get_end_iter(), True)
        self.connect('key-press-event', self.onKeyPress)

    def write(self, text, editable=False):
        gobject.idle_add(self._write, text, editable)

    def _write(self, text, editable=False):
        '''
        Write given text to buffer.

        @param text: Text to append.
        @type text: string
        @param editable: If true, added text is editable.
        @type editable: boolean
        '''
        segments = self.color_pat.split(text)
        segment = segments.pop(0)
        start_mark = self.text_buffer.create_mark(None,
                                                  self.text_buffer.get_end_iter(),
                                                  True)
        self.text_buffer.insert(self.text_buffer.get_end_iter(), segment)

        if segments:
            ansi_tags = self.color_pat.findall(text)
            for tag in ansi_tags:
                i = segments.index(tag)
                self.text_buffer.insert_with_tags_by_name(self.text_buffer.get_end_iter(),
                                                     segments[i+1], tag)
                segments.pop(i)
        if not editable:
            self.text_buffer.apply_tag_by_name('notouch',
                                               self.text_buffer.get_iter_at_mark(start_mark),
                                               self.text_buffer.get_end_iter())
        self.text_buffer.delete_mark(start_mark)
        self.scroll_mark_onscreen(self.mark)


    def showPrompt(self, prompt):
        gobject.idle_add(self._showPrompt, prompt)

    def _showPrompt(self, prompt):
        '''
        Prints prompt at start of line.

        @param prompt: Prompt to print.
        @type prompt: string
        '''
        self._write(prompt)
        self.text_buffer.move_mark(self.line_start,
                                   self.text_buffer.get_end_iter())

    def changeLine(self, text):
        gobject.idle_add(self._changeLine, text)

    def _changeLine(self, text):
        '''
        Replace currently entered command line with given text.

        @param text: Text to use as replacement.
        @type text: string
        '''
        iter = self.text_buffer.get_iter_at_mark(self.line_start)
        iter.forward_to_line_end()
        self.text_buffer.delete(self.text_buffer.get_iter_at_mark(self.line_start), iter)
        self._write(text, True)

    def getCurrentLine(self):
        '''
        Get text in current command line.

        @return: Text of current command line.
        @rtype: string
        '''
        rv = self.text_buffer.get_slice(
          self.text_buffer.get_iter_at_mark(self.line_start),
          self.text_buffer.get_end_iter(), False)
        return rv

    def showReturned(self, text):
        gobject.idle_add(self._showReturned, text)

    def _showReturned(self, text):
        '''
        Show returned text from last command and print new prompt.

        @param text: Text to show.
        @type text: string
        '''
        iter = self.text_buffer.get_iter_at_mark(self.line_start)
        iter.forward_to_line_end()
        self.text_buffer.apply_tag_by_name(
          'notouch',
          self.text_buffer.get_iter_at_mark(self.line_start),
          iter)
        self._write('\n'+text)
        if text:
            self._write('\n')
        self._showPrompt(self.prompt)
        self.text_buffer.move_mark(self.line_start,self.text_buffer.get_end_iter())
        self.text_buffer.place_cursor(self.text_buffer.get_end_iter())

    def onKeyPress(self, widget, event):
        '''
        Key press callback used for correcting behavior for console-like
        interfaces. For example 'home' should go to prompt, not to begining of
        line.

        @param widget: Widget that key press accored in.
        @type widget: gtk.Widget
        @param event: Event object
        @type event: gtk.gdk.Event

        @return: Return True if event should not trickle.
        @rtype: boolean
        '''
        insert_mark = self.text_buffer.get_insert()
        insert_iter = self.text_buffer.get_iter_at_mark(insert_mark)
        selection_mark = self.text_buffer.get_selection_bound()
        selection_iter = self.text_buffer.get_iter_at_mark(selection_mark)
        start_iter = self.text_buffer.get_iter_at_mark(self.line_start)
        if event.keyval == gtk.keysyms.Home:
            if event.state == 0:
                self.text_buffer.place_cursor(start_iter)
                return True
            elif event.state == gtk.gdk.SHIFT_MASK:
                self.text_buffer.move_mark(insert_mark, start_iter)
                return True
        elif event.keyval == gtk.keysyms.Left:
            insert_iter.backward_cursor_position()
            if not insert_iter.editable(True):
                return True
        elif not event.string:
            pass
        elif start_iter.compare(insert_iter) <= 0 and \
              start_iter.compare(selection_iter) <= 0:
            pass
        elif start_iter.compare(insert_iter) > 0 and \
              start_iter.compare(selection_iter) > 0:
            self.text_buffer.place_cursor(start_iter)
        elif insert_iter.compare(selection_iter) < 0:
            self.text_buffer.move_mark(insert_mark, start_iter)
        elif insert_iter.compare(selection_iter) > 0:
            self.text_buffer.move_mark(selection_mark, start_iter)

        return self.onKeyPressExtend(event)

    def onKeyPressExtend(self, event):
        '''
        For some reason we can't extend onKeyPress directly (bug #500900).
        '''
        pass

class IPythonView(ConsoleView, IterableIPShell):
    '''
    Sub-class of both modified IPython shell and L{ConsoleView} this makes
    a GTK+ IPython console.
    '''
    def __init__(self):
        '''
        Initialize. Redirect I/O to console.
        '''
        ConsoleView.__init__(self)
        self.cout = StringIO()
        IterableIPShell.__init__(self, cout=self.cout,cerr=self.cout,
                                 input_func=self.raw_input)
#    self.connect('key_press_event', self.keyPress)
        self.execute()
        self.cout.truncate(0)
        self.showPrompt(self.prompt)
        self.interrupt = False

    def raw_input(self, prompt=''):
        '''
        Custom raw_input() replacement. Get's current line from console buffer.

        @param prompt: Prompt to print. Here for compatability as replacement.
        @type prompt: string

        @return: The current command line text.
        @rtype: string
        '''
        if self.interrupt:
            self.interrupt = False
            raise KeyboardInterrupt
        return self.getCurrentLine()

    def onKeyPressExtend(self, event):
        '''
        Key press callback with plenty of shell goodness, like history,
        autocompletions, etc.

        @param widget: Widget that key press occured in.
        @type widget: gtk.Widget
        @param event: Event object.
        @type event: gtk.gdk.Event

        @return: True if event should not trickle.
        @rtype: boolean
        '''
        if event.state & gtk.gdk.CONTROL_MASK and event.keyval == 99:
            self.interrupt = True
            self._processLine()
            return True
        elif event.keyval == gtk.keysyms.Return:
            self._processLine()
            return True
        elif event.keyval == gtk.keysyms.Up:
            self.changeLine(self.historyBack())
            return True
        elif event.keyval == gtk.keysyms.Down:
            self.changeLine(self.historyForward())
            return True
        elif event.keyval == gtk.keysyms.Tab:
            if not self.getCurrentLine().strip():
                return False
            completed, possibilities = self.complete(self.getCurrentLine())
            if len(possibilities) > 1:
                slice = self.getCurrentLine()
                self.write('\n')
                for symbol in possibilities:
                    self.write(symbol+'\n')
                self.showPrompt(self.prompt)
            self.changeLine(completed or slice)
            return True

    def _processLine(self):
        '''
        Process current command line.
        '''
        self.history_pos = 0
        self.execute()
        rv = self.cout.getvalue()
        if rv: rv = rv.strip('\n')
        self.showReturned(rv)
        self.cout.truncate(0)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Common widgets"""

from gettext import gettext as _
import os

import gnomevfs
import gobject
import gtk
import pango

from kiwi.ui import dialogs

from cf.backends import DBConnectError, DBConnection
from cf.datasources import DatasourceInfo
from cf.ui import GladeWidget


class ConnectionButton(GladeWidget):
    """Connection chooser used in toolbars.

    It is always bound to an SQLEditor.
    """

    def __init__(self, app, xml):
        GladeWidget.__init__(self, app, xml, "tb_connection")
        self._setup_widget()
        self.set_editor(None)
        self.app.datasources.connect("datasource-modified",
                                     lambda *a: self.rebuild_menu())
        self.app.datasources.connect("datasource-added",
                                     lambda *a: self.rebuild_menu())
        self.app.datasources.connect("datasource-deleted",
                                     lambda *a: self.rebuild_menu())

    def _setup_widget(self):
        lbl = gtk.Label("Not connected")
        lbl.set_max_width_chars(15)
        lbl.set_ellipsize(pango.ELLIPSIZE_END)
        lbl.set_use_markup(True)
        lbl.show_all()
        self.set_label_widget(lbl)
        self._label = lbl
        self._menu = gtk.Menu()
        self.set_menu(self._menu)

    def on_new_connection(self, item, datasource_info):
        try:
            conn = datasource_info.dbconnect()
            self._editor.set_connection(conn)
            self.set_editor(self._editor)
        except DBConnectError, err:
            dialogs.error(_(u"Connection failed"), str(err))

    def on_manage_connections(self, *args):
        self.manage_connections()

    def on_set_connection(self, item, connection):
        self._editor.set_connection(connection)
        self.set_editor(self._editor)

    def manage_connections(self):
        """Displays a dialog to manage connections."""
        dlg = ConnectionsDialog(self.app)
        dlg.run()
        dlg.destroy()

    def rebuild_menu(self):
        """Rebuilds the drop-down menu."""
        while self._menu.get_children():
            self._menu.remove(self._menu.get_children()[0])
        if not self._editor:
            return
        ghas_connections = False
        dinfos = self.app.datasources.get_all()
        dinfos.sort(lambda x, y: cmp(x.get_label().lower(),
                                     y.get_label().lower()))
        for datasource_info in dinfos:
            item = gtk.MenuItem(datasource_info.get_label())
            item.show()
            self._menu.append(item)
            menu = gtk.Menu()
            has_connections = False
            for conn in datasource_info.get_connections():
                yitem = gtk.MenuItem((_(u"Connection")+" #%s"
                                      % conn.conn_number))
                yitem.connect("activate", self.on_set_connection, conn)
                yitem.show()
                menu.append(yitem)
                has_connections = True
                ghas_connections = True
            if has_connections:
                sep = gtk.SeparatorMenuItem()
                sep.show()
                menu.append(sep)
            xitem = gtk.MenuItem(_(u"New connection"))
            xitem.connect("activate", self.on_new_connection, datasource_info)
            xitem.show()
            menu.append(xitem)
            item.set_submenu(menu)
        sep = gtk.SeparatorMenuItem()
        sep.show()
        self._menu.append(sep)
        item = gtk.MenuItem(_(u"Show connections"))
        item.connect("activate", lambda *a: self.manage_connections())
        item.show()
        self._menu.append(item)

    def set_editor(self, editor):
        """Associates an editor.

        Arguments:
            editor: SQLEditor instance.
        """
        self._editor = editor
        self.set_sensitive(bool(editor))
        self.rebuild_menu()
        if editor and editor.connection:
            self._label.set_text(editor.connection.get_label())
            markup = ("<b>%s</b>\n%s #%s"
                      % (editor.connection.datasource_info.get_label(),
                         _(u"Connection"), editor.connection.conn_number))
            self.set_tooltip_markup(markup)
        else:
            self._label.set_text("<"+_(u"Not connected")+">")
            self.set_tooltip_markup(_(u"Click to open a connection"))


class DataExportDialog(gtk.FileChooserDialog):
    """Export dialog.

    A modified gtk.FileChooserDialog for exporting data.

    Usage example
    =============

        .. sourcecode:: python

            >>> from cf.ui.widgets import DataExportDialog
            >>> import gtk
            >>> data = [["foo", 1, True], ["bar", 2, False]]
            >>> selected = [2,]
            >>> statement = ("select name, anumber, anotherfield "
                              from foo limit 3;"
            >>> description = (("name", str, None, None, None, None, None),
            ...          ("anumber", int, None, None, None, None, None),
            ...          ("anotherfield", bool, None, None, None, None, None))
            >>> dlg = DataExportDialog(app, instance.widget,
            ...                        data, selected, statement, description)
            >>> if dlg.run() == gtk.RESPONSE_OK:
            ...     dlg.hide()
            ...     dlg.export_data()
            ...
            >>> dlg.destroy()

    """

    def __init__(self, app, parent, data, selected, statement, description):
        """
        The constructor of this class takes 6 arguments:

        :Parameter:
            app
                `CFApplication`_ instance
            parent
                The parent widget, usualy something like
                ``self.instance.widget``
            data
                List of rows (``[ [col1, col2, col3], ...]``)
            selected
                List of indices of selected rows (``None`` means that no
                rows are selected)
            statement
                The SQL statement that produced the data.
                This parameter is only used by some filters to give additional
                information.
            description
                A DB-API2-like description.
                Read the comments on the ``description`` attribute of cursor
                objects in `PEP 249`_ for details.


        .. Note:: Usually there's no need to define ``data`` and
          ``description``
            by hand. If it's a DB-API2-based backend, these parameters are
            retrieved from the cursor object (``cursor.fetchall()`` and
            ``cursor.description``).

        .. _CFApplication: cf.app.CFApplication.html
        .. _PEP 249: http://www.python.org/dev/peps/pep-0249/
        """
        gtk.FileChooserDialog.__init__(self, _(u"Export data"),
                                       parent,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        self.data = data
        self.selected = selected
        self.statement = statement
        self.description = description
        self.app = app
        self._setup_widget()
        self._setup_connections()

    def _setup_widget(self):
        cfg = self.app.config
        vbox = gtk.VBox()
        vbox.set_spacing(5)
        self.set_extra_widget(vbox)
        self.edit_export_options = gtk.CheckButton(_(u"_Edit export options"))
        vbox.pack_start(self.edit_export_options, False, False)
        self.export_selection = gtk.CheckButton(_(u("_Only export "
                                                    "selected rows")))
        self.export_selection.set_sensitive(bool(self.selected))
        vbox.pack_start(self.export_selection, False, False)
        self.open_file = gtk.CheckButton(_(u"_Open file when finished"))
        vbox.pack_start(self.open_file, False, False)
        vbox.show_all()
        if cfg.get("editor.export.recent_folder"):
            self.set_current_folder(cfg.get("editor.export.recent_folder"))
        self._setup_filter()

    def _setup_filter(self):
        cfg = self.app.config
        recent_filter = None
        from cf.plugins.core import PLUGIN_TYPE_EXPORT
        for plugin in self.app.plugins.get_plugins(PLUGIN_TYPE_EXPORT, True):
            filter = gtk.FileFilter()
            filter.set_name(plugin.file_filter_name)
            for pattern in plugin.file_filter_pattern:
                filter.add_pattern(pattern)
            for mime in plugin.file_filter_mime:
                filter.add_mime_type(mime)
            self.add_filter(filter)
            filter.set_data("plugin", plugin)
            if cfg.get("editor.export.recent_filter", None) == plugin.id:
                recent_filter = filter
        if recent_filter:
            self._filter_changed(recent_filter)
        else:
            self._filter_changed(self.get_filter())

    def _setup_connections(self):
        self.connect("notify::filter", self.on_filter_changed)

    def on_filter_changed(self, dialog, param):
        gobject.idle_add(self._filter_changed, self.get_filter())

    def _filter_changed(self, filter):
        plugin = filter.get_data("plugin")
        self.edit_export_options.set_sensitive(plugin.has_options)

    def export_data(self):
        """Exports the data

        This method handles the export options given in the dialog and
        calls the ``export`` method of the choosen `export filter`_.

        .. _export filter: cf.plugins.core.ExportPlugin.html
        """
        self.app.config.set("editor.export.recent_folder",
                            self.get_current_folder())
        plugin = self.get_filter().get_data("plugin")
        self.app.config.set("editor.export.recent_filter", plugin.id)
        if self.export_selection.get_property("sensitive") \
        and self.export_selection.get_active():
            rows = []
            for i in self.selected:
                rows.append(self.data[i])
        else:
            rows = self.data
        opts = {"filename": self.get_filename(),
                "uri": self.get_uri(),
                "query": self.statement}
        if self.edit_export_options.get_property("sensitive") \
        and self.edit_export_options.get_active():
            opts.update(plugin.show_options(self.description, rows))
        plugin.export(self.description, rows, opts)
        if self.open_file.get_active():
            mime = gnomevfs.get_mime_type(opts["uri"])
            app_desc = gnomevfs.mime_get_default_application(mime)
            cmd = app_desc[2].split(" ")
            os.spawnvp(os.P_NOWAIT, cmd[0], cmd+[opts["uri"]])


class ProgressDialog(GladeWidget):
    """Progress dialog with a message

    A simple window with a message, an icon and a progress bar.

    Usage example
    =============

        .. sourcecode:: python

            >>> from cf.ui.widgets import ProgressDialog
            >>> dlg = ProgressDialog(app)
            >>> dlg.show_all()
            >>> dlg = ProgressDialog(app)
            >>> dlg.set_modal(False) # only needed to run it in CF's shell
            >>> dlg.show_all()
            >>> dlg.set_info("This is an information")
            >>> dlg.set_error("Uuups... an error occured")
            >>> dlg.set_progress(0.75)
            >>> dlg.set_finished(True)

    """

    def __init__(self, app, parent=None):
        GladeWidget.__init__(self, app, "crunchyfrog", "progressdialog")
        if parent:
            self.reparent(parent)

    def on_close(self, *args):
        self.destroy()

    def set_progress(self, fraction):
        """Sets fraction for progress bar

        :Parameter:
            fraction
                The fraction (``float``)
        """
        self.xml.get_widget("progress_progress").set_fraction(fraction)

    def set_error(self, message):
        """Displays an error

        :Parameter:
            message
                Error message
        """
        message = gobject.markup_escape_text(message)
        self.xml.get_widget("progress_label").set_markup("<b>"+message+"</b>")
        self.xml.get_widget("progress_image").set_from_stock(
            "gtk-dialog-error", gtk.ICON_SIZE_DIALOG)

    def set_info(self, message):
        """Displays an information

        :Parameter:
            message
                A message
        """
        message = gobject.markup_escape_text(message)
        self.xml.get_widget("progress_label").set_markup("<b>"+message+"</b>")
        self.xml.get_widget("progress_image").set_from_stock(
            "gtk-dialog-info", gtk.ICON_SIZE_DIALOG)

    def set_finished(self, finished):
        """Activates/deactivates close button

        :Parameter:
            finished
                If ``True`` the close button gets sensitive.
        """
        self.xml.get_widget("progress_btn_close").set_sensitive(finished)


class ConnectionsWidget(GladeWidget):
    """Lists datasources and active connections"""

    def __init__(self, app, xml="crunchyfrog"):
        GladeWidget.__init__(self, app, xml, "connections_widget")
        self.refresh()

    def _setup_widget(self):
        self.list_conn = self.xml.get_widget("list_connections")
        model = gtk.TreeStore(gobject.TYPE_PYOBJECT, str)
        model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        self.list_conn.set_model(model)
        col = gtk.TreeViewColumn("", gtk.CellRendererText(), text=1)
        self.list_conn.append_column(col)

    def _setup_connections(self):
        sel = self.list_conn.get_selection()
        sel.connect("changed", self.on_selection_changed)
        self.app.datasources.connect("datasource-added",
                                     self.on_datasource_added)
        self.app.datasources.connect("datasource-deleted",
                                     self.on_datasource_deleted)
        self.app.datasources.connect("datasource-modified",
                                     self.on_datasource_modified)

    def on_connect(self, *args):
        sel = self.list_conn.get_selection()
        model, iter = sel.get_selected()
        if not iter:
            return
        obj = model.get_value(iter, 0)
        if not isinstance(obj, DatasourceInfo):
            return
        conn = obj.dbconnect()
        conn.connect("closed", self.on_connection_closed)
        iter = model.get_iter_first()
        while iter:
            if model.get_value(iter, 0) == conn.datasource_info:
                citer = model.append(iter)
                model.set(citer, 0, conn, 1, conn.get_label())
                break
            iter = model.iter_next(iter)

    def on_disconnect(self, *args):
        sel = self.list_conn.get_selection()
        model, iter = sel.get_selected()
        if not iter:
            return
        obj = model.get_value(iter, 0)
        if not isinstance(obj, DBConnection):
            return
        model.remove(iter)
        obj.close()

    def on_connection_closed(self, connection):
        model = self.list_conn.get_model()
        if not model:
            return
        iter = model.get_iter_first()
        while iter:
            if model.get_value(iter, 0) == connection.datasource_info:
                citer = model.iter_children(iter)
                while citer:
                    if model.get_value(citer, 0) == connection:
                        model.remove(citer)
                        break
                    citer = model.iter_next(citer)
            iter = model.iter_next(iter)

    def on_datasource_added(self, manager, datasource_info):
        model = self.list_conn.get_model()
        iter = model.append(None)
        model.set(iter, 0, datasource_info, 1, datasource_info.get_label())

    def on_datasource_modified(self, manager, datasource_info):
        model = self.list_conn.get_model()
        if not model:
            return
        iter = model.get_iter_first()
        while iter:
            if model.get_value(iter, 0) == datasource_info:
                model.set(iter,
                          0, datasource_info,
                          1, datasource_info.get_label())
                return
            iter = model.iter_next(iter)

    def on_datasource_deleted(self, manager, datasource_info):
        model = self.list_conn.get_model()
        iter = model.get_iter_first()
        while iter:
            if model.get_value(iter, 0) == datasource_info:
                model.remove(iter)
                return
            iter = model.iter_next(iter)

    def on_selection_changed(self, selection):
        model, iter = selection.get_selected()
        is_connection = False
        is_ds = False
        if iter:
            obj = model.get_value(iter, 0)
            is_ds = isinstance(obj, DatasourceInfo)
            is_connection = isinstance(obj, DBConnection)
        self.xml.get_widget("btn_disconnect").set_sensitive(is_connection)
        self.xml.get_widget("btn_connect").set_sensitive(is_ds)

    def refresh(self):
        """Initializes the data model"""
        model = self.list_conn.get_model()
        for datasource_info in self.app.datasources.get_all():
            iter = model.append(None)
            model.set(iter, 0, datasource_info, 1, datasource_info.get_label())
            for conn in datasource_info.get_connections():
                citer = model.append(iter)
                model.set(citer, 0, conn, 1, conn.get_label())


class ConnectionsDialog(GladeWidget):
    """Dialog displaying connections"""

    def __init__(self, app):
        GladeWidget.__init__(self, app, "crunchyfrog", "connectionsdialog")

    def _setup_widget(self):
        self.connections = ConnectionsWidget(self.app, self.xml)


class CustomImageMenuItem(gtk.ImageMenuItem):
    """Menu item with custom image.

    This widget simplifies the creation of an gtk.ImageMenuItem with an
    custom image. icon_name is used to lookup an icon in the default GTK
    icon theme and so it is not restricted to stock id's. The additional method
    set_markup() can be used to set a markup string as the menu items
    label.

    Note: It is not recommended to use set_markup() because it is very
      unusual to have formatted text in menu items.
    """

    def __init__(self, icon_name, label, is_markup=False):
        """
        The constructor of this class takes up to 3 arguments:

        :Parameter:
            icon_name
                Name of the icon or stock id
            label
                Menu item label
            is_markup
                If ``True``, ``label`` is set as markup (default: ``False``)
        """
        gtk.ImageMenuItem.__init__(self, "gtk-open")
        self.set_icon_name(icon_name)
        if is_markup:
            self.set_markup(label)
        else:
            self.set_label(label)

    def set_icon_name(self, icon_name):
        """Sets the image

        Args:
          icon_name: Icon name of the image.
        Raises:
          GError: if the icon isn't present in icon theme.
        """
        it = gtk.icon_theme_get_default()
        pb = it.load_icon(icon_name, gtk.ICON_SIZE_MENU,
                          gtk.ICON_LOOKUP_FORCE_SVG)
        img = gtk.image_new_from_pixbuf(pb)
        self.set_image(gtk.image_new_from_pixbuf(pb))

    def set_label(self, label):
        """Sets the label

        :Parameter:
            label
                Menu item label
        """
        self.get_children()[0].set_label(label)

    def set_markup(self, markup):
        """Sets the label as markup

        :Parameter:
            markup
                Menu item markup
        """
        self.get_children()[0].set_markup(markup)



# Copyright (C) 2008 Andi Albrecht, albrecht.andi@gmail.com
#
# This module is part of python-sqlparse and is released under
# the BSD License: http://www.opensource.org/licenses/bsd-license.php.

"""Exceptions raised by sqlparse."""


class SQLParseError(Exception):
    """Base error for all errors in this module."""


class SQLParseParseError(SQLParseError):
    """Raised when SQL parsing fails."""


class SQLParseMultiStatementError(SQLParseError):
    """Raised when sqlparse() is called with a multi-statement string."""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Core objects"""

import gobject
import gnomevfs
import sys
import os
from inspect import isclass
import imp
import zipimport
from kiwi.ui import dialogs

import logging
log = logging.getLogger("PLUGINS")

from gettext import gettext as _

PLUGIN_TYPE_GENERIC = 0
PLUGIN_TYPE_BACKEND = 1
PLUGIN_TYPE_EXPORT = 2
PLUGIN_TYPE_EDITOR = 3

from cf import USER_PLUGIN_DIR, PLUGIN_DIR, USER_PLUGIN_URI

from cf.plugins.mixins import InstanceMixin, MenubarMixin, EditorMixin
from cf.plugins.mixins import UserDBMixin
from cf.ui.widgets import ProgressDialog

class GenericPlugin(gobject.GObject):
    """Plugin base class"""

    id = None
    name = None
    description = None
    long_description = None
    icon = None
    author = None
    license = None
    homepage = None
    version = None
    has_custom_options = False
    plugin_type = PLUGIN_TYPE_GENERIC
    INIT_ERROR = None

    def __init__(self, app):
        """
        The constructor of this class takes one argument:

        :Parameter:
            app
                `CFApplication`_ instance

        .. _CFApplication: cf.app.CFApplication.html
        """
        self.app = app
        self.__gobject_init__()

    @classmethod
    def run_custom_options_dialog(cls, app):
        """Runs a preferences dialog

        If ``has_custom_options`` is ``True`` this method will
        be called if the user clicks on the *Configure plugin* button
        in the preferences dialog.

        :Parameter:
            app
                `CFApplication`_ instance

        .. _CFApplication: cf.app.CFApplication.html
        """
        pass

    def shutdown(self):
        """Called when the plugin is deactivated."""
        pass

class ExportPlugin(GenericPlugin):
    """Export filter base class"""
    icon = "gtk-save-as"
    file_filter_name = None
    file_filter_mime = []
    file_filter_pattern = []
    has_options = False
    plugin_type = PLUGIN_TYPE_EXPORT

    def __init__(self, app):
        GenericPlugin.__init__(self, app)

    def export(self, description, rows, options=dict()):
        raise NotImplementedError

    def show_options(self, description, rows):
        return dict()


class DBBackendPlugin(GenericPlugin):
    """Database backend base class"""

    icon = "stock_database"
    context_help_pattern = None
    plugin_type = PLUGIN_TYPE_BACKEND
    password_option = None # key for options which holds the password

    def __init__(self, app):
        GenericPlugin.__init__(self, app)
        self.schema = None
        self.reference = None

    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        return data_widgets, []
    # TODO: Documentation

    @classmethod
    def get_label(cls, datasource_info):
        s = "%s@%s on %s" % (datasource_info.options.get("user", None) or "??",
                             datasource_info.options.get("database", None) or "??",
                             datasource_info.options.get("host", None) or "??")
        if datasource_info.name:
            s = '%s (%s)' % (datasource_info.name, s)
        return s

    def test_connection(self, data):
        raise NotImplementedError

    def password_prompt(self):
        return dialogs.password(_(u"Password required"), _(u"Enter the password to connect to this database."))

# Key: plugin type
# Value: 2-tuple (label, expected class)
PLUGIN_TYPES_MAP = {
    PLUGIN_TYPE_GENERIC : (_(u"Miscellaneous"), GenericPlugin),
    PLUGIN_TYPE_BACKEND : (_(u"Database backends"), DBBackendPlugin),
    PLUGIN_TYPE_EXPORT : (_(u"Export filter"), ExportPlugin),
    PLUGIN_TYPE_EDITOR : (_(u"Editor"), GenericPlugin),
}

class PluginManager(gobject.GObject):
    """Plugin manager

    An instance of this class is accessible through the ``plugins``
    attribute of an `CFApplication`_ instance.

    Signals
    =======

        plugin-added
            ``def callback(manager, plugin, user_param1, ...)``

            Emitted when a plugin was added to the registry.

        plugin-removed
            ``def callback(manager, plugin, user_param1, ...)``

            Emitted when a plugin is removed from the registry.

        plugin-active
            ``def callback(manager, plugin, active, user_param1, ...)``

            Emitted when a plugin is activated or deactivated. `active`
            is either ``True`` or ``False``.

    .. _CFApplication: cf.app.CFApplication.html
    """

    plugin_types = PLUGIN_TYPES_MAP

    __gsignals__ = {
        "plugin-added" : (gobject.SIGNAL_RUN_LAST,
                          gobject.TYPE_NONE,
                          (gobject.TYPE_PYOBJECT,)),
        "plugin-removed" : (gobject.SIGNAL_RUN_LAST,
                            gobject.TYPE_NONE,
                            (gobject.TYPE_PYOBJECT,)),
        "plugin-active" : (gobject.SIGNAL_RUN_LAST,
                           gobject.TYPE_NONE,
                           (gobject.TYPE_PYOBJECT, bool)),
    }

    def __init__(self, app):
        """
        The constructor of this class takes one argument:

        :Parameter:
            app
                `CFApplication`_ instance

        .. _CFApplication: cf.app.CFApplication.html
        """
        self.app = app
        self.__gobject_init__()
        self.__plugins = dict()
        self.__active_plugins = dict()
        gnomevfs.monitor_add(USER_PLUGIN_DIR, gnomevfs.MONITOR_DIRECTORY, self.on_plugin_folder_changed)
        gnomevfs.monitor_add(PLUGIN_DIR, gnomevfs.MONITOR_DIRECTORY, self.on_plugin_folder_changed)
        self.app.register_shutdown_task(self.on_app_shutdown, _(u"Closing plugins"))
        self.app.cb.connect("instance-created", self.on_instance_created)
        self.refresh()
        self._first_run()

    def on_app_shutdown(self):
        for plugin in self.__active_plugins.values():
            plugin.shutdown()

    def on_instance_created(self, cb, instance):
        for plugin in self.__active_plugins.values():
            if isinstance(plugin, InstanceMixin):
                self.init_instance_mixins(plugin, instance)

    def on_plugin_folder_changed(self, folder, path, change):
        if change in [gnomevfs.MONITOR_EVENT_DELETED, gnomevfs.MONITOR_EVENT_CREATED]:
            gobject.idle_add(self.refresh)

    def _first_run(self):
        if not self.app.options.first_run:
            return
        log.info("Activating default backends...")
        from cf.plugins.core import PLUGIN_TYPE_BACKEND
        for plugin in self.get_plugins(PLUGIN_TYPE_BACKEND):
            self.set_active(plugin, True)

    def get_plugins(self, plugin_type, active_only=False):
        """Returns a list of plugins.

        :Parameter:
            plugin_type
                a ``PLUGIN_TYPE_*`` constant

            active_only
                If set to ``True`` only activated plugins are returned.

        :Returns:
            List of `plugins`_

        .. _plugins: cf.plugins.core.GenericPlugin.html
        """
        ret = list()
        if active_only:
            plugins = self.__active_plugins.values()
        else:
            plugins = self.__plugins.values()
        for item in plugins:
            if item.plugin_type == plugin_type:
                ret.append(item)
        return ret

    def _get_modules(self, path):
        modules = []
        if path not in sys.path:
            sys.path.insert(0, path)
        for item in os.listdir(path):
            if item.startswith(".") or item.startswith("_"):
                continue
            name, ext = os.path.splitext(item)
            if ext and ext not in [".zip", ".py"]:
                continue
            if ext == ".zip":
                sys.path.insert(0, os.path.join(path, item))
                importer = zipimport.zipimporter(os.path.join(path, item))
                importer = importer.find_module(name)
                if not importer:
                    continue
                mod = importer.load_module(name)
                modules.append(mod)
            else:
                try:
                    modinfo = imp.find_module(name)
                except ImportError:
                    continue
                try:
                    mod = imp.load_module(name, *modinfo)
                except Exception:
                    print str(sys.exc_info()[1])
                    try: del sys.modules[name]
                    except KeyError: pass
                    continue
                modules.append(mod)
        return modules

    def _get_plugins(self, module):
        plugins = []
        for name in dir(module):
            obj = getattr(module, name)
            if isclass(obj) and issubclass(obj, GenericPlugin) \
            and obj not in [GenericPlugin, ExportPlugin, DBBackendPlugin]:
                plugins.append(obj)
        return plugins

    def refresh(self):
        """Refreshs the plugin registry.

        This method is called when the contents of a plugin folder
        changes.
        """
        from cf.plugins import builtin
        modules = [builtin]
        for path in [PLUGIN_DIR, USER_PLUGIN_DIR]:
            modules += self._get_modules(path)
        plugins = []
        for module in modules:
            plugins += self._get_plugins(module)
        ids_found = []
        for plugin in plugins:
            if not self.__plugins.has_key(plugin.id):
                self.__plugins[plugin.id] = plugin
                l = self.app.config.get("plugins.active", [])
                if plugin.id in l and not plugin.INIT_ERROR:
                    self.set_active(plugin, True)
                elif plugin.id in l and plugin.INIT_ERROR:
                    self.set_active(plugin, False)
                self.emit("plugin-added", plugin)
            ids_found.append(plugin.id)
        for id, plugin in self.__plugins.items():
            if id not in ids_found:
                l = self.app.config.get("plugins.active", [])
                if id in l:
                    self.set_active(plugin, False)
                del self.__plugins[id]
                self.emit("plugin-removed", plugin)

    def set_active(self, plugin, active, instance=None):
        """Activates / deactivates a plugin

        :Parameter:
            plugin
                Plugin to activate / deactivate
            active
                If ``True`` the plugin gets activated, otherwise it will
                be deactivated.
        """
        id = None
        for key, value in self.__plugins.items():
            if value == plugin:
                id = key
                break
        if not id:
            return
        l = self.app.config.get("plugins.active", [])
        if active:
            x = plugin(self.app)
            self.__active_plugins[plugin] = x
            if isinstance(x, UserDBMixin):
                x.userdb_set(self.app.userdb)
                x.userdb_init()
            if isinstance(x, InstanceMixin):
                for instance in self.app.get_instances():
                    self.init_instance_mixins(x, instance)
            if id not in l:
                l.append(id)
        elif not active:
            x = self.__active_plugins.get(plugin, None)
            if x:
                if isinstance(x, InstanceMixin):
                    for instance in self.app.get_instances():
                        self.unload_instance_mixins(x, instance)
                x.shutdown()
                del self.__active_plugins[plugin]
            if id in l:
                l.remove(id)
        self.app.config.set("plugins.active", l)
        self.emit("plugin-active", plugin, active)

    def is_active(self, plugin):
        """Returns ``True`` if the plugin is active

        :Parameter:
            plugin
                A plugin

        :Returns: ``True`` if the plugin is active.
        """
        if isinstance(plugin, GenericPlugin):
            plugin = plugin.__class__
        return self.__active_plugins.has_key(plugin)

    def by_id(self, id, active_only=True):
        """Returns a plugin by its id

        :Parameter:
            id
                Plugin ID
            active_only
                If ``True`` only active plugins are returned.

        :Returns: `Plugin`_ or ``None``

        .. _Plugin: cf.plugins.core.GenricPlugin.html
        """
        plugins = self.__active_plugins.values()
        if not active_only:
            plugins += self.__plugins.values()
        for plugin in plugins:
            if plugin.id == id:
                return plugin
        return None

    def init_instance_mixins(self, plugin, instance):
        plugin.init_instance(instance)
        if isinstance(plugin, MenubarMixin):
            plugin.menubar_load(instance.xml.get_widget("menubar"), instance)
        if isinstance(plugin, EditorMixin):
            self.editor_notify(instance._editor, instance)

    def unload_instance_mixins(self, plugin, instance):
        if isinstance(plugin, MenubarMixin):
            plugin.menubar_unload(instance.xml.get_widget("menubar"), instance)

    def editor_notify(self, editor, instance):
        """Called by an instance when the current editor has changed

        :Parameter:
            editor
                an editor or ``None``
            instance
                an instance
        """
        for plugin in self.__active_plugins.values():
            if isinstance(plugin, EditorMixin):
                plugin.set_editor(editor, instance)

    def install_plugin(self, uri):
        """Installs a plugin from URI

        :Parameter:
            uri
                URI pointing to .egg file
        """
        def progress_cb(info, dlg):
            if info.bytes_total:
                fraction = info.total_bytes_copied/float(info.bytes_total)
            else:
                fraction = 0
            dlg.set_progress(fraction)
            return True
        if uri is None:
            return
        source = gnomevfs.URI(uri)
        dest = gnomevfs.URI(USER_PLUGIN_URI).append_file_name(source.short_name)
        try:
            dlg = ProgressDialog(self.app)
            dlg.show_all()
            dlg.set_info(_(u"Copying files..."))
            gnomevfs.xfer_uri(source, dest, gnomevfs.XFER_DEFAULT,
                              gnomevfs.XFER_ERROR_MODE_QUERY,
                              gnomevfs.XFER_OVERWRITE_MODE_QUERY,
                              progress_cb, dlg)
            dlg.set_info(_(u"Plugin installed."))
        except:
            err = sys.exc_info()[1]
            dlg.set_error(str(err))
        dlg.set_finished(True)
